@@ -2,9 +2,10 @@
  ** OptiPNG: Advanced PNG optimization program.
  ** http://optipng.sourceforge.net/
  **
- ** Copyright (C) 2001-2008 Cosmin Truta.
- ** OptiPNG is open-source software, and is distributed under the same
- ** licensing and warranty terms as libpng.
+ ** Copyright (C) 2001-2009 Cosmin Truta.
+ **
+ ** This software is distributed under the zlib license.
+ ** Please see the attached LICENSE for more information.
  **
  ** PNG optimization is described in detail in the PNG-Tech article
  ** "A guide to PNG optimization"
@@ -17,12 +18,11 @@
  ** the pngrewrite program by Jason Summers.
  **
  ** Requirements:
- **    ANSI C or ISO C compiler and library.
- **    POSIX library for enhanced functionality.
- **    zlib version 1.2.1 or newer (version 1.2.3 is bundled).
- **    libpng version 1.2.9 or newer (version 1.2.33 is bundled).
- **    pngxtern (version 0.6.2 is bundled).
- **    cexcept (version 2.0.1 is bundled).
+ **    ANSI C89, ISO C90 or ISO C99 compiler and library.
+ **    zlib version 1.2.1 or newer.
+ **    libpng version 1.2.9 or newer, with pngxtern.
+ **    cexcept version 2.0.1 or newer.
+ **    POSIX or Windows API for enhanced functionality.
  **/
 
 #include <ctype.h>
@@ -56,23 +56,19 @@ static const char *msg_license =
    "\n";
 
 static const char *msg_short_help =
-   "Type \"optipng -h\" for advanced help.\n"
-   "\n"
    "Synopsis:\n"
    "    optipng [options] files ...\n"
    "Files:\n"
    "    Image files of type: PNG, BMP, GIF, PNM or TIFF\n"
    "Basic options:\n"
-   "    -h, -help\t\tshow the advanced help\n"
-   "    -v\t\t\tverbose mode / show copyright, version and build info\n"
-   "    -o  <level>\t\toptimization level (0-7)\t\tdefault 2\n"
-   "    -i  <type>\t\tinterlace type (0-1)\t\t\tdefault <input>\n"
-   "    -k, -keep\t\tkeep a backup of the modified files\n"
-   "    -q, -quiet\t\tquiet mode\n"
+   "    -?, -h, -help\tshow the extended help\n"
+   "    -o <level>\t\toptimization level (0-7)\t\tdefault 2\n"
+   "    -v\t\t\tverbose mode / show copyright and version info\n"
    "Examples:\n"
    "    optipng file.png\t\t\t(default speed)\n"
    "    optipng -o5 file.png\t\t(moderately slow)\n"
-   "    optipng -o7 file.png\t\t(very slow)\n";
+   "    optipng -o7 file.png\t\t(very slow)\n"
+   "Type \"optipng -h\" for extended help.\n";
 
 static const char *msg_help =
    "Synopsis:\n"
@@ -80,41 +76,19 @@ static const char *msg_help =
    "Files:\n"
    "    Image files of type: PNG, BMP, GIF, PNM or TIFF\n"
    "Basic options:\n"
-   "    -h, -help\t\tshow this help\n"
-   "    -v\t\t\tverbose mode / show copyright, version and build info\n"
-   "    -o  <level>\t\toptimization level (0-7)\t\tdefault 2\n"
-#if 0  /* not implemented */
-   "    -b  <depth>\t\tbit depth (1,2,4,8,16)\t\t\tdefault <min>\n"
-   "    -c  <type>\t\tcolor type (0,2,3,4,6)\t\t\tdefault <input>\n"
-#endif
-   "    -i  <type>\t\tinterlace type (0-1)\t\t\tdefault <input>\n"
-   "    -k, -keep\t\tkeep a backup of the modified files\n"
-   "    -q, -quiet\t\tquiet mode\n"
-   "Advanced options:\n"
-   "    -zc <levels>\tzlib compression levels (1-9)\t\tdefault 9\n"
-   "    -zm <levels>\tzlib memory levels (1-9)\t\tdefault 8\n"
-   "    -zs <strategies>\tzlib compression strategies (0-3)\tdefault 0-3\n"
-#ifdef WBITS_8_OK
-   "    -zw <window size>\tzlib window size (32k,16k,8k,4k,2k,1k,512,256)\n"
-#else
-   "    -zw <window size>\tzlib window size (32k,16k,8k,4k,2k,1k,512)\n"
-#endif
-   "    -f  <filters>\tPNG delta filters (0-5)\t\t\tdefault 0,5\n"
-   "    -nb\t\t\tno bit depth reduction\n"
-   "    -nc\t\t\tno color type reduction\n"
-   "    -np\t\t\tno palette reduction\n"
-#if 0  /* metadata optimization is not implemented */
-   "    -nm\t\t\tno metadata optimization\n"
-#endif
-   "    -nz\t\t\tno IDAT recompression (also disable reductions)\n"
+   "    -?, -h, -help\tshow this help\n"
+   "    -o <level>\t\toptimization level (0-7)\t\tdefault 2\n"
+   "    -v\t\t\tverbose mode / show copyright and version info\n"
+   "General options:\n"
    "    -fix\t\tenable error recovery\n"
    "    -force\t\tenforce writing of a new output file\n"
-   "    -full\t\tproduce a full report on IDAT (might reduce speed)\n"
 #if 0  /* parallel processing is not implemented */
    "    -jobs <number>\tallow parallel jobs\n"
 #endif
+   "    -keep\t\tkeep a backup of the modified files\n"
    "    -preserve\t\tpreserve file attributes if possible\n"
-   "    -simulate\t\trun in simulation mode, do not create output files\n"
+   "    -quiet\t\tquiet mode\n"
+   "    -simulate\t\tsimulation mode\n"
    "    -snip\t\tcut one image out of multi-image or animation files\n"
 #if 0  /* multi-image splitting is not implemented */
    "    -split\t\tsplit multi-image/animation files into separate images\n"
@@ -123,7 +97,26 @@ static const char *msg_help =
    "    -dir <directory>\twrite output file(s) to <directory>\n"
    "    -log <file>\t\tlog messages to <file>\n"
    "    --\t\t\tstop option switch parsing\n"
-   "Optimization parameters:\n"
+   "Optimization options:\n"
+#if 0  /* not implemented */
+   "    -b  <depth>\t\tbit depth (1,2,4,8,16)\t\t\tdefault <min>\n"
+   "    -c  <type>\t\tcolor type (0,2,3,4,6)\t\t\tdefault <input>\n"
+#endif
+   "    -f  <filters>\tPNG delta filters (0-5)\t\t\tdefault 0,5\n"
+   "    -i  <type>\t\tPNG interlace type (0-1)\t\tdefault <input>\n"
+   "    -zc <levels>\tzlib compression levels (1-9)\t\tdefault 9\n"
+   "    -zm <levels>\tzlib memory levels (1-9)\t\tdefault 8\n"
+   "    -zs <strategies>\tzlib compression strategies (0-3)\tdefault 0-3\n"
+   "    -zw <window size>\tzlib window size (32k,16k,8k,4k,2k,1k,512,256)\n"
+   "    -full\t\tproduce a full report on IDAT (might reduce speed)\n"
+   "    -nb\t\t\tno bit depth reduction\n"
+   "    -nc\t\t\tno color type reduction\n"
+   "    -np\t\t\tno palette reduction\n"
+#if 0  /* metadata optimization is not implemented */
+   "    -nm\t\t\tno metadata optimization\n"
+#endif
+   "    -nz\t\t\tno IDAT recompression (also disable reductions)\n"
+   "Optimization details:\n"
    "    The optimization level presets\n"
    "        -o0  <=>  -nz\n"
    "        -o1  <=>  [use the libpng heuristics]\t(1 trial)\n"
@@ -147,7 +140,9 @@ static const char *msg_help =
 
 static enum { OP_NONE, OP_HELP, OP_RUN } operation;
 static struct opng_options options;
-static FILE *logfile;
+static FILE *con_file;
+static FILE *log_file;
+int start_of_line;
 
 
 /** Error handling **/
@@ -226,9 +221,9 @@ str2long(const char *str, long *value)
 
 /** Command line error handling **/
 static void
-error_option(const char *opt_desc, const char *opt_arg)
+err_option(const char *opt_desc, const char *opt_arg)
 {
-    /* Issue an error regarding the incorrect use of the option, and exit. */
+    /* Issue an error regarding the incorrect use of the option. */
     if (opt_arg != NULL && opt_arg[0] != 0)
         error("Invalid %s: %s", opt_desc, opt_arg);
     else
@@ -336,18 +331,6 @@ parse_args(int argc, char *argv[])
         {
             options.help = 1;
         }
-        else if (strcmp(opt, "v") == 0)
-        {
-            options.ver = 1;
-        }
-        else if (string_prefix_min_cmp("keep", opt, 1) == 0)
-        {
-            options.keep = 1;
-        }
-        else if (string_prefix_min_cmp("quiet", opt, 1) == 0)
-        {
-            options.quiet = 1;
-        }
         else if (string_prefix_min_cmp("fix", opt, 2) == 0)
         {
             options.fix = 1;
@@ -360,17 +343,9 @@ parse_args(int argc, char *argv[])
         {
             options.full = 1;
         }
-        else if (string_prefix_min_cmp("preserve", opt, 2) == 0)
+        else if (string_prefix_min_cmp("keep", opt, 1) == 0)
         {
-            options.preserve = 1;
-        }
-        else if (string_prefix_min_cmp("simulate", opt, 2) == 0)
-        {
-            options.simulate = 1;
-        }
-        else if (string_prefix_min_cmp("snip", opt, 2) == 0)
-        {
-            options.snip = 1;
+            options.keep = 1;
         }
         else if (strcmp(opt, "nb") == 0)
         {
@@ -380,12 +355,12 @@ parse_args(int argc, char *argv[])
         {
             options.nc = 1;
         }
+#if 0
         else if (strcmp(opt, "nm") == 0)
         {
-#if 0
             options.nm = 1;
-#endif
         }
+#endif
         else if (strcmp(opt, "np") == 0)
         {
             options.np = 1;
@@ -394,6 +369,35 @@ parse_args(int argc, char *argv[])
         {
             options.nz = 1;
         }
+        else if (string_prefix_min_cmp("preserve", opt, 1) == 0)
+        {
+            options.preserve = 1;
+        }
+        else if (string_prefix_min_cmp("quiet", opt, 1) == 0)
+        {
+            options.quiet = 1;
+        }
+        else if (string_prefix_min_cmp("simulate", opt, 2) == 0)
+        {
+            options.simulate = 1;
+        }
+        else if (string_prefix_min_cmp("snip", opt, 2) == 0)
+        {
+            options.snip = 1;
+        }
+        else if (strcmp(opt, "v") == 0)
+        {
+            options.verbose = 1;
+            options.version = 1;
+        }
+        else if (string_prefix_min_cmp("verbose", opt, 4) == 0)
+        {
+            options.verbose = 1;
+        }
+        else if (string_prefix_min_cmp("version", opt, 4) == 0)
+        {
+            options.version = 1;
+        }
         else  /* possibly an option with an argument */
         {
             if (xopt == NULL)
@@ -417,7 +421,7 @@ parse_args(int argc, char *argv[])
         else if (strcmp(opt, "o") == 0)
         {
             if (str2long(xopt, &lval) != 0 || lval < 0 || lval > 99)
-                error_option("optimization level", xopt);
+                err_option("optimization level", xopt);
             val = (int)lval;
             if (options.optim_level < 0)
                 options.optim_level = val;
@@ -427,7 +431,7 @@ parse_args(int argc, char *argv[])
         else if (strcmp(opt, "i") == 0)
         {
             if (str2long(xopt, &lval) != 0 || lval < 0 || lval > 1)
-                error_option("interlace type", xopt);
+                err_option("interlace type", xopt);
             val = (int)lval;
             if (options.interlace < 0)
                 options.interlace = (int)val;
@@ -447,25 +451,25 @@ parse_args(int argc, char *argv[])
         else if (strcmp(opt, "f") == 0)
         {
             if (bitset_parse(xopt, &set) != 0)
-                error_option("filter(s)", xopt);
+                err_option("filter(s)", xopt);
             options.filter_set |= set;
         }
         else if (strcmp(opt, "zc") == 0)
         {
             if (bitset_parse(xopt, &set) != 0)
-                error_option("zlib compression level(s)", xopt);
+                err_option("zlib compression level(s)", xopt);
             options.compr_level_set |= set;
         }
         else if (strcmp(opt, "zm") == 0)
         {
             if (bitset_parse(xopt, &set) != 0)
-                error_option("zlib memory level(s)", xopt);
+                err_option("zlib memory level(s)", xopt);
             options.mem_level_set |= set;
         }
         else if (strcmp(opt, "zs") == 0)
         {
             if (bitset_parse(xopt, &set) != 0)
-                error_option("zlib compression strategy", xopt);
+                err_option("zlib compression strategy", xopt);
             options.strategy_set |= set;
         }
         else if (strcmp(opt, "zw") == 0)
@@ -476,7 +480,7 @@ parse_args(int argc, char *argv[])
                 if ((1L << val) == lval)
                     break;
             if (val < 8)
-                error_option("zlib window size", xopt);
+                err_option("zlib window size", xopt);
             if (options.window_bits == 0)
                 options.window_bits = val;
             else if (options.window_bits != val)
@@ -487,26 +491,26 @@ parse_args(int argc, char *argv[])
             if (options.out_name != NULL)
                 error("Duplicate output file name");
             if (xopt[0] == 0)
-                error_option("output file name", xopt);
+                err_option("output file name", xopt);
             options.out_name = xopt;
         }
-        else if (string_prefix_min_cmp("dir", opt, 2) == 0)
+        else if (string_prefix_min_cmp("dir", opt, 1) == 0)
         {
             if (options.dir_name != NULL)
                 error("Duplicate output dir name");
             if (xopt[0] == 0)
-                error_option("output dir name", xopt);
+                err_option("output dir name", xopt);
             options.dir_name = xopt;
         }
-        else if (string_prefix_min_cmp("log", opt, 2) == 0)
+        else if (string_prefix_min_cmp("log", opt, 1) == 0)
         {
             if (options.log_name != NULL)
                 error("Duplicate log file name");
             if (xopt[0] == 0)
-                error_option("log file name", xopt);
+                err_option("log file name", xopt);
             options.log_name = xopt;
         }
-        else if (string_prefix_min_cmp("jobs", opt, 2) == 0)
+        else if (string_prefix_min_cmp("jobs", opt, 1) == 0)
         {
             error("Parallel processing is not implemented");
         }
@@ -542,14 +546,19 @@ parse_args(int argc, char *argv[])
 static void
 app_init(void)
 {
-    setvbuf(stdout, NULL, _IONBF, 0);
+    /* Initialize the console output. */
+    con_file = (!options.quiet || options.help) ? stdout : NULL;
+
+    /* Open the log file, line-buffered, if requested. */
     if (options.log_name != NULL)
     {
-        /* Open the log file, line-buffered. */
-        if ((logfile = fopen(options.log_name, "a")) == NULL)
+        if ((log_file = fopen(options.log_name, "a")) == NULL)
             error("Can't open log file: %s\n", options.log_name);
-        setvbuf(logfile, NULL, _IOLBF, BUFSIZ);
+        setvbuf(log_file, NULL, _IOLBF, BUFSIZ);
     }
+
+    /* Initialize the internal printing routines. */
+    start_of_line = 1;
 }
 
 
@@ -557,10 +566,10 @@ app_init(void)
 static void
 app_finish(void)
 {
-    if (logfile != NULL)
+    if (log_file != NULL)
     {
         /* Close the log file. */
-        fclose(logfile);
+        fclose(log_file);
     }
 }
 
@@ -571,46 +580,86 @@ app_printf(const char *fmt, ...)
 {
     va_list arg_ptr;
 
-    /* Print w/ logging. */
-    if (!options.quiet)
+    if (fmt[0] == 0)
+        return;
+    start_of_line = (fmt[strlen(fmt) - 1] == '\n') ? 1 : 0;
+
+    if (con_file != NULL)
     {
         va_start(arg_ptr, fmt);
-        vfprintf(stdout, fmt, arg_ptr);
+        vfprintf(con_file, fmt, arg_ptr);
         va_end(arg_ptr);
     }
-    if (logfile != NULL)
+    if (log_file != NULL)
     {
-        /* "\r"            -->  reset line in console, new line in log */
-        /* "[:blank:]*\r"  -->  wipe line in console, leave log intact */
-        if (strcmp(fmt, "\r") == 0)
-            fmt = "\n";
-        else if (fmt[0] != '\0' && fmt[strlen(fmt) - 1] == '\r')
-            return;
         va_start(arg_ptr, fmt);
-        vfprintf(logfile, fmt, arg_ptr);
+        vfprintf(log_file, fmt, arg_ptr);
         va_end(arg_ptr);
     }
 }
 
 
-/** Application-defined flush callback **/
+/** Application-defined control print callback **/
 static void
-app_flush(void)
+app_print_cntrl(int cntrl_code)
 {
-    if (!options.quiet)
-        fflush(stdout);
-    if (logfile != NULL)
-        fflush(logfile);
+    const char *con_str, *log_str;
+    int i;
+
+    if (cntrl_code == '\r')
+    {
+        /* CR: reset line in console, new line in log file. */
+        con_str = "\r";
+        log_str = "\n";
+        start_of_line = 1;
+    }
+    else if (cntrl_code == '\v')
+    {
+        /* VT: new line if current line is not empty, nothing otherwise. */
+        if (!start_of_line)
+        {
+            con_str = log_str = "\n";
+            start_of_line = 1;
+        }
+        else
+            con_str = log_str = "";
+    }
+    else if (cntrl_code < 0 && cntrl_code > -80 && start_of_line)
+    {
+        /* Minus N: erase first N characters from line, in console only. */
+        if (con_file != NULL)
+        {
+            for (i = 0; i > cntrl_code; --i)
+                fputc(' ', con_file);
+        }
+        con_str = "\r";
+        log_str = "";
+    }
+    else
+    {
+        /* Unhandled control code (due to internal error): show err marker. */
+        con_str = log_str = "<?>";
+    }
+
+    if (con_file != NULL)
+        fprintf(con_file, con_str);
+    if (log_file != NULL)
+        fprintf(log_file, log_str);
 }
 
 
-/** Application-defined progress-bar callback **/
+/** Application-defined progress update callback **/
 static void
-app_progress(unsigned long num, unsigned long denom)
+app_progress(unsigned long current_step, unsigned long total_steps)
 {
+    /* There will be a potentially long wait, so flush the console output. */
+    if (con_file != NULL)
+        fflush(con_file);
+    /* An eager flush of the line-buffered log file is not very important. */
+
     /* A GUI application would normally update a progress bar. */
-    /* We do nothing in this console program. */
-    if (num && denom)
+    /* Here we ignore the progress info. */
+    if (current_step && total_steps)
         return;
 }
 
@@ -624,10 +673,10 @@ process_files(int argc, char *argv[])
     int i;
 
     /* Initialize the optimization engine. */
-    ui.printf_fn   = app_printf;
-    ui.flush_fn    = app_flush;
-    ui.progress_fn = app_progress;
-    ui.panic_fn    = panic;
+    ui.printf_fn      = app_printf;
+    ui.print_cntrl_fn = app_print_cntrl;
+    ui.progress_fn    = app_progress;
+    ui.panic_fn       = panic;
     if (opng_initialize(&options, &ui) != 0)
         panic("Can't initialize optimization engine");
 
@@ -655,18 +704,17 @@ main(int argc, char *argv[])
 {
     int result;
 
-    /* Parse the command options. */
+    /* Parse the user options and initialize the application. */
     parse_args(argc, argv);
-
-    /* Initialize the application. */
     app_init();
 
-    /* Print the program/library version, copyright and licensing info. */
+    /* Print the copyright and version info. */
     app_printf("%s", msg_intro);
-    if (options.ver)
+    if (options.version)
     {
+        /* Print the licensing and extended version info. */
         app_printf(msg_license);
-        app_printf("Compiled with libpng version %s and zlib version %s\n\n",
+        app_printf("Using libpng version %s and zlib version %s\n\n",
             png_get_libpng_ver(NULL), zlibVersion());
         /* Print the help text only if explicitly requested. */
         if (operation == OP_HELP && !options.help)
